
#pragma once

// Genode includes
#include <base/env.h>
#include <base/heap.h>

// protobuf includes
#include <google/protobuf/message.h>
#include <google/protobuf/util/time_util.h>

// gapfruit includes
#include <tee_session/connection.h>
#include <tee/packet_allocator.h>

// generated includes
#include <tee.pb.h>

// stdcxx includes
#include <functional>


namespace Tee {

	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::Signal_context_capability;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::Message;
	using google::protobuf::Timestamp;
	using google::protobuf::util::TimeUtil;
	using gapfruit::protobuf::tee::CounterRequest;

	class Connection_asynchronous;

	struct Proto_callbacks;
}


struct Tee::Proto_callbacks
{
	std::function<void(Timestamp)> counter_request;
	std::function<void(Timestamp)> execute_init_request;
	std::function<void(Timestamp)> execute_request;
	std::function<void(Timestamp)> quote_request;
};


class Tee::Connection_asynchronous
{
	private:

		using Signal_handler = Genode::Signal_handler<Connection_asynchronous>;

		Env&                       _env;
		Heap                       _heap             { _env.ram(), _env.rm() };
		Allocator_avl              _tx_block_alloc   { &_heap };
		Tee::Connection            _connection;
		char const*                _session_label;
		const Proto_callbacks&     _callbacks;

		// signal handlers for connection
		Signal_handler             _rx_data          { _env.ep(), *this, &Connection_asynchronous::_handle_rx_data };
		Signal_handler             _rx_ack           { _env.ep(), *this, &Connection_asynchronous::_handle_rx_ack };
		Signal_handler             _tx_data          { _env.ep(), *this, &Connection_asynchronous::_handle_tx_data };
		Signal_handler             _tx_ack           { _env.ep(), *this, &Connection_asynchronous::_handle_tx_ack };

		Signal_context_capability  _counter_request  { };

		// methods used by the signal handlers
		void _handle_rx_data();
		void _handle_rx_ack();
		void _handle_tx_data();
		void _handle_tx_ack();

		Timestamp _timestamp_from_system_time();

		// dummy method to initialize callback when the user
		// desn't want asynchronous data transfers.
		Proto_callbacks            _callback_dummies { [] (Timestamp) {},
		                                               [] (Timestamp) {},
		                                               [] (Timestamp) {},
		                                               [] (Timestamp) {} };

	public:

		enum { DEFAULT_BUF_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE * 128 };

		struct Serialize_failed    : Genode::Exception { };
		struct Deserialize_failed  : Genode::Exception { };
		struct Transmission_error  : Genode::Exception { };

		Connection_asynchronous(Genode::Env& env, char const* label,
		                        size_t tx_buf_size = DEFAULT_BUF_SIZE,
		                        size_t rx_buf_size = DEFAULT_BUF_SIZE) :
			_env           { env },
			_connection    { env, &_tx_block_alloc, tx_buf_size, rx_buf_size, label },
			_session_label { label },
			_callbacks     { _callback_dummies }
		{ }

		~Connection_asynchronous() = default;

		/**
		 * Sends a protobuf message
		 *
		 * \param msg  The protobuf message to send.
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 * \result     Tee execution result
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		Result send(const google::protobuf::MessageLite& msg, const Tee::Session_id sid);

		/**
		 * Send a CounterResponse
		 *
		 * \param  value      counter value
		 * \param  timestamp  timestamp when the message was received
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		void send_counter_response(const uint64_t value, const Timestamp& timestamp);
};
