
#pragma once

// Genode includes
#include <base/env.h>
#include <base/heap.h>

// protobuf includes
#include <google/protobuf/message.h>
#include <google/protobuf/util/time_util.h>

// gapfruit includes
#include <tee_session/connection.h>
#include <tee/packet_allocator.h>

// generated includes
#include <tee.pb.h>

// stdcxx includes
#include <functional>


namespace Tee {

	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::Signal_context_capability;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::Message;
	using google::protobuf::Timestamp;
	using google::protobuf::util::TimeUtil;
	using gapfruit::protobuf::tee::CounterRequest;

	class Connection_synchronous;
}


class Tee::Connection_synchronous
{
	private:

		using Signal_handler = Genode::Signal_handler<Connection_synchronous>;

		Env&                       _env;
		Heap                       _heap             { _env.ram(), _env.rm() };
		Allocator_avl              _tx_block_alloc   { &_heap };
		Tee::Connection            _connection;
		char const*                _session_label;

		// methods used by the signal handlers
		void _handle_rx_data();
		void _handle_rx_ack();
		void _handle_tx_data();
		void _handle_tx_ack();

	public:

		enum { DEFAULT_BUF_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE * 128 };

		struct Serialize_failed    : Genode::Exception { };
		struct Deserialize_failed  : Genode::Exception { };
		struct Transmission_error  : Genode::Exception { };

		Connection_synchronous(Genode::Env& env, char const* label,
		                       size_t tx_buf_size = DEFAULT_BUF_SIZE,
		                       size_t rx_buf_size = DEFAULT_BUF_SIZE) :
			_env           { env },
			_connection    { env, &_tx_block_alloc, tx_buf_size, rx_buf_size, label },
			_session_label { label }
		{ }

		~Connection_synchronous() = default;

		/**
		 * Sends a protobuf message
		 *
		 * \param msg  The protobuf message to send.
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 * \result     Tee execution result
		 *
		 * \throw Transmission_error  E.g. when there is no sink to send to.
		 * \throw Serialize_failed    Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed Could not allocate packet on connection.
		 */
		Result send(const google::protobuf::MessageLite& msg, const Tee::Session_id sid);

		/**
		 * Blocking receive of a protobuf message
		 *
		 * \param msg  The protobuf message to receive
		 * \param sid  The session id. Sinks need to reply using the received
		 *             session id so the source can track sessions.
		 *
		 * \throw Deserialize_failed  Protobuf message could not be deserialized
		 */
		void receive(google::protobuf::MessageLite& msg, Tee::Session_id& sid);
};
